
-- | Encodings for nucleotides and pairs. Currently very ViennaRNA-centric.
-- This might change over time.

{-# LANGUAGE NoMonomorphismRestriction, TypeSynonymInstances, FlexibleInstances, TypeFamilies, StandaloneDeriving, MultiParamTypeClasses #-}
module Biobase.RNA where

import Data.Char (toUpper)
import Data.Ix
import Control.Monad (liftM)
import qualified Data.Vector.Unboxed as V
import qualified Data.Vector.Unboxed.Base as VB
import qualified Data.Vector.Primitive as P
import qualified Data.Vector.Generic         as G
import qualified Data.Vector.Generic.Mutable as M



-- | single nucleotide encoding

data Nucleotide = E | A | C | G | U
  deriving (Show,Read,Eq,Ord,Enum,Bounded,Ix)



-- | transform characters into nucleotide representation

char2Nucleotide c = f $ toUpper c where
  f 'A' = A
  f 'C' = C
  f 'G' = G
  f 'U' = U
  f _   = E



-- | transform list of characters into list of nucleotides

string2Nucleotides = map char2Nucleotide

-- | transform list of nucleotides into list of characters

nucleotides2String = concatMap show






-- * basic definition of primary sequence and conversion to

type Primary = V.Vector Nucleotide

class AsPrimary a where
  {-# INLINE asPrimary #-}
  asPrimary :: a -> Primary

instance AsPrimary String where
  asPrimary = V.fromList . string2Nucleotides

instance AsPrimary [Nucleotide] where
  asPrimary = V.fromList



-- | Primary back to strings

primary2String :: Primary -> String
primary2String = concatMap show . V.toList



-- | Vienna RNA style pair encodings using single tokens.

data ViennaPair = NP | CG | GC | GU | UG | AU | UA | NS
  deriving (Show, Read, Eq, Ord, Enum, Bounded, Ix)



-- * convenience lists

acgu = [A .. U]

acguStr = map show acgu

acguPairs = [(b1, b2) | b1 <- acgu, b2 <- acgu]

eacgu = [E .. U]

npnsP = [NP .. NS]

cguaP = [CG .. UA]



-- * pair transformation


-- {{{ transformation from and to pairs

{-# INLINE toPair #-}
toPair :: Nucleotide -> Nucleotide -> ViennaPair
toPair C G = CG
toPair G C = GC
toPair G U = GU
toPair U G = UG
toPair A U = AU
toPair U A = UA
toPair _ _ = NP


{-# INLINE pairs #-}
pairs :: Nucleotide -> Nucleotide -> Bool
pairs a b = (>NP) $ toPair a b

vienna2Tuple CG = (C,G)
vienna2Tuple GC = (G,C)
vienna2Tuple GU = (G,U)
vienna2Tuple UG = (U,G)
vienna2Tuple AU = (A,U)
vienna2Tuple UA = (U,A)
vienna2Tuple _  = error "non-standard pairs can't be backcasted"

vienna2RevTuple CG = (G,C)
vienna2RevTuple GC = (C,G)
vienna2RevTuple GU = (U,G)
vienna2RevTuple UG = (G,U)
vienna2RevTuple AU = (U,A)
vienna2RevTuple UA = (A,U)
vienna2RevTuple _  = error "non-standard pairs can't be backcasted"

tuple2Vienna (C,G) = CG
tuple2Vienna (G,C) = GC
tuple2Vienna (G,U) = GU
tuple2Vienna (U,G) = UG
tuple2Vienna (A,U) = AU
tuple2Vienna (U,A) = UA
tuple2Vienna _     = NS

tuple2RevVienna (a,b) = tuple2Vienna (b,a)

-- }}}



-- {{{ Data.Vector.Unboxed instance

-- | instance for Data.Vector.Unboxed

newtype instance VB.MVector s Nucleotide = MV_Nucleotide (P.MVector s Int)
newtype instance VB.Vector    Nucleotide = V_Nucleotide  (P.Vector    Int)

instance VB.Unbox Nucleotide

instance M.MVector V.MVector Nucleotide where
  {-# INLINE basicLength #-}
  {-# INLINE basicUnsafeSlice #-}
  {-# INLINE basicOverlaps #-}
  {-# INLINE basicUnsafeNew #-}
  {-# INLINE basicUnsafeNewWith #-}
  {-# INLINE basicUnsafeRead #-}
  {-# INLINE basicUnsafeWrite #-}
  {-# INLINE basicClear #-}
  {-# INLINE basicSet #-}
  {-# INLINE basicUnsafeCopy #-}
  {-# INLINE basicUnsafeGrow #-}
  basicLength (MV_Nucleotide v) = M.basicLength v
  basicUnsafeSlice i n (MV_Nucleotide v) = MV_Nucleotide $ M.basicUnsafeSlice i n v
  basicOverlaps (MV_Nucleotide v1) (MV_Nucleotide v2) = M.basicOverlaps v1 v2
  basicUnsafeNew n = MV_Nucleotide `liftM` M.basicUnsafeNew n
  basicUnsafeNewWith n x = MV_Nucleotide `liftM` M.basicUnsafeNewWith n (fromEnum x)
  basicUnsafeRead (MV_Nucleotide v) i = toEnum `liftM` M.basicUnsafeRead v i
  basicUnsafeWrite (MV_Nucleotide v) i x = M.basicUnsafeWrite v i (fromEnum x)
  basicClear (MV_Nucleotide v) = M.basicClear v
  basicSet (MV_Nucleotide v) x = M.basicSet v (fromEnum x)
  basicUnsafeCopy (MV_Nucleotide v1) (MV_Nucleotide v2) = M.basicUnsafeCopy v1 v2
  basicUnsafeGrow (MV_Nucleotide v) n = MV_Nucleotide `liftM` M.basicUnsafeGrow v n

instance G.Vector V.Vector Nucleotide where
  {-# INLINE unsafeFreeze #-}
  {-# INLINE basicLength #-}
  {-# INLINE basicUnsafeSlice #-}
  {-# INLINE basicUnsafeIndexM #-}
  {-# INLINE elemseq #-}
  unsafeFreeze (MV_Nucleotide v) = V_Nucleotide `liftM` G.unsafeFreeze v
  basicLength (V_Nucleotide v) = G.basicLength v
  basicUnsafeSlice i n (V_Nucleotide v) = V_Nucleotide $ G.basicUnsafeSlice i n v
  basicUnsafeIndexM (V_Nucleotide v) i = toEnum `liftM` G.basicUnsafeIndexM v i
  elemseq _ = seq

-- }}}



-- TODO if more pair encodings arrive, then put toPair and isPair into a type
-- class and derive for each encoding.
